import inspect
from functools import wraps
from typing import Any, Callable, List, Optional, Type, Union, cast

from django.http import HttpRequest, HttpResponse
from ninja.signature import is_async

from ninja_extra import exceptions
from ninja_extra.conf import settings
from ninja_extra.constants import THROTTLED_FUNCTION
from ninja_extra.controllers import ControllerBase, RouteContext
from ninja_extra.dependency_resolver import service_resolver

from .model import BaseThrottle


def throttle(*func_or_throttle_classes: Any, **init_kwargs: Any) -> Callable[..., Any]:
    isfunction = (
        inspect.isfunction(func_or_throttle_classes[0])
        if len(func_or_throttle_classes) == 1
        else False
    )

    if isfunction:
        func = func_or_throttle_classes[0]
        throttle_classes: List[Type[BaseThrottle]] = settings.THROTTLE_CLASSES
        return _inject_throttling(func, *throttle_classes, **init_kwargs)

    def wrapper(view_func: Callable[..., Any]) -> Any:
        return _inject_throttling(view_func, *func_or_throttle_classes, **init_kwargs)

    return wrapper


def _run_throttles(
    *throttle_classes: Type[BaseThrottle],
    request_or_controller: Union[HttpRequest, ControllerBase],
    response: Optional[HttpResponse] = None,
    **init_kwargs: Any,
) -> None:
    """
    Run all throttles for a request.
    Raises an appropriate exception if the request is throttled.
    """

    request = cast(
        HttpRequest,
        (
            request_or_controller.context.request  # type:ignore
            if isinstance(request_or_controller, ControllerBase)
            else request_or_controller
        ),
    )

    throttle_durations = []

    for throttle_class in throttle_classes:
        throttling: BaseThrottle = throttle_class(**init_kwargs)
        if not throttling.allow_request(request):
            # Filter out `None` values which may happen in case of config / rate
            duration = throttling.wait()
            if duration is not None:
                throttle_durations.append(duration)

    if throttle_durations:
        duration = max(throttle_durations, default=None)
        raise exceptions.Throttled(duration)


def _inject_throttling(
    func: Callable[..., Any],
    *throttle_classes: Type[BaseThrottle],
    **init_kwargs: Any,
) -> Callable[..., Any]:
    setattr(func, THROTTLED_FUNCTION, True)
    if is_async(func):
        return _async_inject_throttling_handler(func, *throttle_classes, **init_kwargs)
    return _sync_inject_throttling_handler(func, *throttle_classes, **init_kwargs)


def _sync_inject_throttling_handler(
    func: Callable[..., Any],
    *throttle_classes: Type[BaseThrottle],
    **init_kwargs: Any,
) -> Callable[..., Any]:
    @wraps(func)
    def as_view(
        request_or_controller: Union[HttpRequest, ControllerBase], *args: Any, **kw: Any
    ) -> Any:
        ctx = cast(Optional[RouteContext], service_resolver(RouteContext))
        _run_throttles(
            *throttle_classes,
            request_or_controller=request_or_controller,
            response=ctx.response if ctx else None,
            **init_kwargs,
        )

        res = func(request_or_controller, *args, **kw)
        return res

    return as_view


def _async_inject_throttling_handler(
    func: Callable[..., Any],
    *throttle_classes: Type[BaseThrottle],
    **init_kwargs: Any,
) -> Callable[..., Any]:
    @wraps(func)
    async def as_view(
        request_or_controller: Union[HttpRequest, ControllerBase], *args: Any, **kw: Any
    ) -> Any:
        ctx = cast(Optional[RouteContext], service_resolver(RouteContext))
        _run_throttles(
            *throttle_classes,
            request_or_controller=request_or_controller,
            response=ctx.response if ctx else None,
            **init_kwargs,
        )

        res = await func(request_or_controller, *args, **kw)
        return res

    return as_view



import datetime
import functools
import re
import typing as t
import uuid
from urllib.parse import parse_qs, urlparse

from django.db.models import Model as DjangoModel
from django.db.models import QuerySet
from ninja import Query, Schema
from ninja.pagination import PaginationBase
from ninja.params import Body, Path
from pydantic import UUID4, create_model
from pydantic import BaseModel as PydanticModel

from ninja_extra.permissions import BasePermission
from ninja_extra.shortcuts import add_ninja_contribute_args

from ... import status
from ...exceptions import NotFound
from ...pagination import PageNumberPaginationExtra, PaginatedResponseSchema, paginate
from ..route import route

if t.TYPE_CHECKING:
    from ..base import ModelControllerBase


# Match parameters in URL paths, eg. '{param}', and '{int:param}'
PARAM_REGEX = re.compile("{([a-zA-Z_][a-zA-Z0-9_]*:)?([a-zA-Z_][a-zA-Z0-9_]*)}")


class PathCompiledResult(t.NamedTuple):
    param_convertors: t.Dict[str, t.Any]
    query_parameters: t.Dict[str, t.Any]

    def has_any_parameter(self) -> bool:
        return len(self.param_convertors) > 0 or len(self.query_parameters) > 0


STRING_TYPES: t.Dict[str, t.Type] = {
    "int": int,
    "path": str,
    "slug": str,
    "str": str,
    "uuid": UUID4,
    "datetime": datetime.datetime,
    "date": datetime.date,
}


def compile_path(path: str) -> PathCompiledResult:
    """
    Given a path string, like: "/{str:username}"

    regex:      "/(?P<username>[^/]+)"
    format:     "/{username}"
    convertors: {str:"username"}
    """
    duplicated_params = set()
    param_convertors = {}

    parsed_url = urlparse(path)
    query_parameters = parse_qs(parsed_url.query)

    for match in PARAM_REGEX.finditer(path):
        convertor_type, param_name = match.groups("str")
        convertor = convertor_type.rstrip(":")

        if param_name in param_convertors:
            duplicated_params.add(param_name)

        param_convertors[param_name] = STRING_TYPES[convertor]

    if duplicated_params:
        names = ", ".join(sorted(duplicated_params))
        ending = "s" if len(duplicated_params) > 1 else ""
        raise ValueError(f"Duplicated param name{ending} {names} at path {path}")

    return PathCompiledResult(param_convertors, query_parameters)


class ModelEndpointFactory:
    @classmethod
    def _clean_path(cls, path: str) -> str:
        working_path = path.split("?")
        return working_path[0]

    @classmethod
    def _change_name(cls, name_prefix: str) -> str:
        unique = str(uuid.uuid4())[:6]
        return f"{name_prefix}_{unique}"

    @classmethod
    def _path_resolver(cls, path: str) -> t.Callable:
        """
        Create a decorator to parse `path` parameters and resolve them during endpoint.
        For example:
        path=/{int:id}/tags/{post_id}?query=int&query1=int
        this will create two path parameters `id` and `post_id` and two query parameters `query` and `query1`
        for the decorated endpoint
        """
        compiled_path = compile_path(path)

        def get_path_fields() -> t.Generator:
            for path_name, path_type in compiled_path.param_convertors.items():
                yield path_name, (path_type, ...)

        def get_query_fields() -> t.Generator:
            for path_name, path_type in compiled_path.query_parameters.items():
                path_type.append("str")
                yield path_name, (STRING_TYPES[path_type[0]], ...)

        def decorator(func: t.Callable) -> t.Callable:
            if not compiled_path.has_any_parameter():
                return func

            _ninja_contribute_args: t.List[t.Tuple] = getattr(
                func, "_ninja_contribute_args", []
            )

            unique_key = str(uuid.uuid4().hex)[:5]

            path_construct_name = f"PathModel{unique_key}"
            query_construct_name = f"QueryModel{unique_key}"

            path_fields = dict(get_path_fields())
            query_fields = dict(get_query_fields())

            if path_fields:
                dynamic_path_model = create_model(
                    path_construct_name, __base__=Schema, **path_fields
                )
                add_ninja_contribute_args(
                    func,
                    (path_construct_name, dynamic_path_model, Path(...)),
                )

            if query_fields:
                dynamic_query_model = create_model(
                    query_construct_name, __base__=Schema, **query_fields
                )

                add_ninja_contribute_args(
                    func,
                    (query_construct_name, dynamic_query_model, Query(...)),
                )

            @functools.wraps(func)
            def path_parameter_resolver(*args: t.Any, **kwargs: t.Any) -> t.Any:
                func_kwargs = dict(kwargs)

                if path_construct_name in func_kwargs:
                    dynamic_model_path_instance = func_kwargs.pop(path_construct_name)
                    func_kwargs.update(dynamic_model_path_instance.dict())

                if query_construct_name in func_kwargs:
                    dynamic_model_query_instance = func_kwargs.pop(query_construct_name)
                    func_kwargs.update(dynamic_model_query_instance.dict())

                return func(*args, **func_kwargs)

            return path_parameter_resolver

        return decorator

    @classmethod
    def create(
        cls,
        schema_in: t.Type[PydanticModel],
        schema_out: t.Type[PydanticModel],
        path: str = "/",
        status_code: int = status.HTTP_201_CREATED,
        url_name: t.Optional[str] = None,
        custom_handler: t.Optional[t.Callable[..., t.Any]] = None,
        description: t.Optional[str] = None,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = "Create new item",
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable:
        """
        Creates a POST Action
        """
        working_path = cls._clean_path(path)

        @route.post(
            working_path,
            response={status_code: schema_out},
            url_name=url_name,
            description=description,
            operation_id=operation_id,
            summary=summary,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            include_in_schema=include_in_schema,
            permissions=permissions,
            openapi_extra=openapi_extra,
        )
        @cls._path_resolver(path)
        def create_item(
            self: "ModelControllerBase",
            data: schema_in = Body(default=...),  # type:ignore[valid-type]
            **kwargs: t.Any,
        ) -> t.Any:
            instance = (
                custom_handler(self, data, **kwargs)
                if custom_handler
                else self.service.create(data, **kwargs)
            )
            assert instance, "`service.create` or  `custom_handler` must return a value"
            return instance

        create_item.__name__ = cls._change_name("create_item")
        return create_item  # type:ignore[no-any-return]

    @classmethod
    def update(
        cls,
        path: str,
        lookup_param: str,
        schema_in: t.Type[PydanticModel],
        schema_out: t.Type[PydanticModel],
        status_code: int = status.HTTP_200_OK,
        url_name: t.Optional[str] = None,
        description: t.Optional[str] = None,
        object_getter: t.Optional[t.Callable[..., DjangoModel]] = None,
        custom_handler: t.Optional[t.Callable[..., t.Any]] = None,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = "Update an item",
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable:
        """
        Creates a PUT Action
        """
        working_path = cls._clean_path(path)

        @route.put(
            working_path,
            response={status_code: schema_out},
            url_name=url_name,
            description=description,
            operation_id=operation_id,
            summary=summary,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            include_in_schema=include_in_schema,
            permissions=permissions,
            openapi_extra=openapi_extra,
        )
        @cls._path_resolver(path)
        def update_item(
            self: "ModelControllerBase",
            data: schema_in = Body(default=...),  # type:ignore[valid-type]
            **kwargs: t.Any,
        ) -> t.Any:
            pk = kwargs.pop(lookup_param)
            obj = (
                object_getter(self, pk=pk, **kwargs)
                if object_getter
                else self.service.get_one(pk=pk, **kwargs)
            )

            if not obj:  # pragma: no cover
                raise NotFound()

            self.check_object_permissions(obj)
            instance = (
                custom_handler(self, instance=obj, schema=data, **kwargs)
                if custom_handler
                else self.service.update(instance=obj, schema=data, **kwargs)
            )
            assert instance, "`service.update` or `custom_handler` must return a value"
            return instance

        update_item.__name__ = cls._change_name("update_item")
        return update_item  # type:ignore[no-any-return]

    @classmethod
    def patch(
        cls,
        path: str,
        lookup_param: str,
        schema_in: t.Type[PydanticModel],
        schema_out: t.Type[PydanticModel],
        status_code: int = status.HTTP_200_OK,
        url_name: t.Optional[str] = None,
        description: t.Optional[str] = None,
        object_getter: t.Optional[t.Callable[..., DjangoModel]] = None,
        custom_handler: t.Optional[t.Callable[..., t.Any]] = None,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = "Patch Item Update",
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable:
        """
        Creates a PATCH Action
        """
        working_path = cls._clean_path(path)

        @route.patch(
            working_path,
            response={status_code: schema_out},
            url_name=url_name,
            description=description,
            operation_id=operation_id,
            summary=summary,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            include_in_schema=include_in_schema,
            permissions=permissions,
            openapi_extra=openapi_extra,
        )
        @cls._path_resolver(path)
        def patch_item(
            self: "ModelControllerBase",
            data: schema_in = Body(default=...),  # type:ignore[valid-type]
            **kwargs: t.Any,
        ) -> t.Any:
            pk = kwargs.pop(lookup_param)
            obj = (
                object_getter(self, pk=pk, **kwargs)
                if object_getter
                else self.service.get_one(pk=pk, **kwargs)
            )
            if not obj:  # pragma: no cover
                raise NotFound()
            self.check_object_permissions(obj)
            instance = (
                custom_handler(self, instance=obj, schema=data, **kwargs)
                if custom_handler
                else self.service.patch(instance=obj, schema=data, **kwargs)
            )
            assert instance, "`service.patch()` or `custom_handler` must return a value"
            return instance

        patch_item.__name__ = cls._change_name("patch_item")
        return patch_item  # type:ignore[no-any-return]

    @classmethod
    def find_one(
        cls,
        path: str,
        lookup_param: str,
        schema_out: t.Type[PydanticModel],
        status_code: int = status.HTTP_200_OK,
        url_name: t.Optional[str] = None,
        description: t.Optional[str] = None,
        object_getter: t.Optional[t.Callable[..., DjangoModel]] = None,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = "Find a specific item",
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable:
        """
        Creates a GET Action
        """
        working_path = cls._clean_path(path)

        @route.get(
            working_path,
            response={status_code: schema_out},
            url_name=url_name,
            description=description,
            operation_id=operation_id,
            summary=summary,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            include_in_schema=include_in_schema,
            permissions=permissions,
            openapi_extra=openapi_extra,
        )
        @cls._path_resolver(path)
        def get_item(self: "ModelControllerBase", **kwargs: t.Any) -> t.Any:
            pk = kwargs.pop(lookup_param)
            obj = (
                object_getter(self, pk=pk, **kwargs)
                if object_getter
                else self.service.get_one(pk=pk, **kwargs)
            )
            if not obj:  # pragma: no cover
                raise NotFound()
            self.check_object_permissions(obj)
            return obj

        get_item.__name__ = cls._change_name("get_item")
        return get_item  # type:ignore[no-any-return]

    @classmethod
    def list(
        cls,
        schema_out: t.Type[PydanticModel],
        path: str = "/",
        status_code: int = status.HTTP_200_OK,
        url_name: t.Optional[str] = None,
        description: t.Optional[str] = None,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = "List of Items",
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
        queryset_getter: t.Optional[t.Callable[..., QuerySet]] = None,
        pagination_response_schema: t.Optional[
            t.Type[PydanticModel]
        ] = PaginatedResponseSchema,
        pagination_class: t.Optional[
            t.Type[PaginationBase]
        ] = PageNumberPaginationExtra,
        **paginate_kwargs: t.Any,
    ) -> t.Callable:
        """
        Creates a GET Action to list Items
        """
        working_path = cls._clean_path(path)

        @cls._path_resolver(path)
        def list_items(self: "ModelControllerBase", **kwargs: t.Any) -> t.Any:
            """List Items of testing"""
            if queryset_getter:
                return queryset_getter(self, **kwargs)
            return self.service.get_all(**kwargs)

        if pagination_response_schema and pagination_class:
            list_items = paginate(pagination_class, **paginate_kwargs)(list_items)
            list_items = route.get(
                working_path,
                response={
                    status_code: pagination_response_schema[schema_out]  # type:ignore[index]
                },
                url_name=url_name,
                description=description,
                operation_id=operation_id,
                summary=summary,
                tags=tags,
                deprecated=deprecated,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                include_in_schema=include_in_schema,
                permissions=permissions,
                openapi_extra=openapi_extra,
            )(list_items)
        else:
            list_items = route.get(
                working_path,
                response={status_code: t.List[schema_out]},  # type:ignore[valid-type]
                url_name=url_name,
                description=description,
                operation_id=operation_id,
                summary=summary,
                tags=tags,
                deprecated=deprecated,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                include_in_schema=include_in_schema,
                permissions=permissions,
                openapi_extra=openapi_extra,
            )(list_items)

        list_items.__name__ = cls._change_name("list_items")
        return list_items  # type:ignore[no-any-return]

    @classmethod
    def delete(
        cls,
        path: str,
        lookup_param: str,
        status_code: int = status.HTTP_204_NO_CONTENT,
        url_name: t.Optional[str] = None,
        description: t.Optional[str] = None,
        object_getter: t.Optional[t.Callable[..., DjangoModel]] = None,
        custom_handler: t.Optional[t.Callable[..., t.Any]] = None,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = None,
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable:
        """
        Creates a DELETE Action to list Items
        """
        working_path = cls._clean_path(path)

        @route.delete(
            working_path,
            url_name=url_name,
            response={status_code: str},
            description=description,
            operation_id=operation_id,
            summary="Delete An Item",
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            include_in_schema=include_in_schema,
            permissions=permissions,
            openapi_extra=openapi_extra,
        )
        @cls._path_resolver(path)
        def delete_item(self: "ModelControllerBase", **kwargs: t.Any) -> t.Any:
            pk = kwargs.pop(lookup_param)
            obj = (
                object_getter(self, pk=pk, **kwargs)
                if object_getter
                else self.service.get_one(pk=pk, **kwargs)
            )
            if not obj:  # pragma: no cover
                raise NotFound()
            self.check_object_permissions(obj)
            custom_handler(
                self, instance=obj, **kwargs
            ) if custom_handler else self.service.delete(instance=obj, **kwargs)
            return self.create_response(message="", status_code=status_code)

        delete_item.__name__ = cls._change_name("delete_item")
        return delete_item  # type:ignore[no-any-return]



import inspect
import logging
from collections import OrderedDict
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

from asgiref.sync import sync_to_async
from django.core.paginator import InvalidPage, Page, Paginator
from django.db.models import QuerySet
from django.http import HttpRequest
from ninja import Schema
from ninja.constants import NOT_SET
from ninja.pagination import LimitOffsetPagination, PageNumberPagination, PaginationBase
from ninja.signature import is_async
from ninja.types import DictStrAny
from pydantic import Field

from ninja_extra.conf import settings
from ninja_extra.exceptions import NotFound
from ninja_extra.schemas import PaginatedResponseSchema
from ninja_extra.shortcuts import add_ninja_contribute_args
from ninja_extra.urls import remove_query_param, replace_query_param

logger = logging.getLogger()

if TYPE_CHECKING:  # pragma: no cover
    from .controllers import ControllerBase

__all__ = [
    "PageNumberPagination",
    "PageNumberPaginationExtra",
    "PaginationBase",
    "LimitOffsetPagination",
    "paginate",
    "PaginatedResponseSchema",
    "PaginatorOperation",
    "AsyncPaginatorOperation",
]


class PageNumberPaginationExtra(PaginationBase):
    class Input(Schema):
        page: int = Field(1, gt=0)
        page_size: int = Field(100, lt=200)

    page_query_param = "page"
    page_size_query_param = "page_size"

    max_page_size = 200
    paginator_class = Paginator

    def __init__(
        self,
        page_size: int = settings.PAGINATION_PER_PAGE,
        max_page_size: Optional[int] = None,
        pass_parameter: Optional[str] = None,
    ) -> None:
        super().__init__(pass_parameter=pass_parameter)
        self.page_size = page_size
        self.max_page_size = max_page_size or 200
        self.Input = self.create_input()  # type:ignore

    def create_input(self) -> Type[Input]:
        class DynamicInput(PageNumberPaginationExtra.Input):
            page: int = Field(1, gt=0)
            page_size: int = Field(self.page_size, lt=self.max_page_size)

        return DynamicInput

    def paginate_queryset(
        self,
        queryset: QuerySet,
        pagination: Input,
        request: Optional[HttpRequest] = None,
        **params: DictStrAny,
    ) -> Any:
        assert request, "request is required"
        current_page_number = pagination.page
        paginator = self.paginator_class(queryset, pagination.page_size)
        try:
            url = request.build_absolute_uri()
            page: Page = paginator.page(current_page_number)
            return self.get_paginated_response(base_url=url, page=page)
        except InvalidPage as exc:  # pragma: no cover
            msg = "Invalid page. {page_number} {message}".format(
                page_number=current_page_number, message=str(exc)
            )
            raise NotFound(msg) from exc

    def get_paginated_response(self, *, base_url: str, page: Page) -> DictStrAny:
        return OrderedDict(
            [
                ("count", page.paginator.count),
                ("next", self.get_next_link(base_url, page=page)),
                ("previous", self.get_previous_link(base_url, page=page)),
                ("results", list(page)),
            ]
        )

    @classmethod
    def get_response_schema(
        cls, response_schema: Union[Type[Schema], Type[Any]]
    ) -> Any:
        return PaginatedResponseSchema[response_schema]  # type: ignore[valid-type]

    def get_next_link(self, url: str, page: Page) -> Optional[str]:
        if not page.has_next():
            return None
        page_number = page.next_page_number()
        return replace_query_param(url, self.page_query_param, page_number)

    def get_previous_link(self, url: str, page: Page) -> Optional[str]:
        if not page.has_previous():
            return None
        page_number = page.previous_page_number()
        if page_number == 1:
            return remove_query_param(url, self.page_query_param)
        return replace_query_param(url, self.page_query_param, page_number)


@overload
def paginate() -> Callable[..., Any]:  # pragma: no cover
    ...


@overload
def paginate(
    func_or_pgn_class: Any = NOT_SET, **paginator_params: Any
) -> Callable[..., Any]:  # pragma: no cover
    ...


def paginate(
    func_or_pgn_class: Any = NOT_SET, **paginator_params: Any
) -> Callable[..., Any]:
    isfunction = inspect.isfunction(func_or_pgn_class)
    isnotset = func_or_pgn_class == NOT_SET

    pagination_class: Type[PaginationBase] = settings.PAGINATION_CLASS

    if isfunction:
        return _inject_pagination(func_or_pgn_class, pagination_class)

    if not isnotset:
        pagination_class = func_or_pgn_class

    def wrapper(func: Callable[..., Any]) -> Any:
        return _inject_pagination(func, pagination_class, **paginator_params)

    return wrapper


def _inject_pagination(
    func: Callable[..., Any],
    paginator_class: Type[PaginationBase],
    **paginator_params: Any,
) -> Callable[..., Any]:
    paginator: PaginationBase = paginator_class(**paginator_params)
    paginator_kwargs_name = "pagination"
    paginator_operation_class = PaginatorOperation
    if is_async(func):
        paginator_operation_class = AsyncPaginatorOperation
    paginator_operation = paginator_operation_class(
        paginator=paginator, view_func=func, paginator_kwargs_name=paginator_kwargs_name
    )

    return paginator_operation.as_view


class PaginatorOperation:
    def __init__(
        self,
        *,
        paginator: PaginationBase,
        view_func: Callable,
        paginator_kwargs_name: str = "pagination",
    ) -> None:
        self.paginator = paginator
        self.paginator_kwargs_name = paginator_kwargs_name
        self.view_func = view_func

        paginator_view = self.get_view_function()
        _ninja_contribute_args: List[Tuple] = getattr(
            self.view_func, "_ninja_contribute_args", []
        )
        paginator_view._ninja_contribute_args = (  # type:ignore[attr-defined]
            _ninja_contribute_args
        )
        add_ninja_contribute_args(
            paginator_view,
            (
                self.paginator_kwargs_name,
                self.paginator.Input,
                self.paginator.InputSource,
            ),
        )
        paginator_view.paginator_operation = self  # type:ignore[attr-defined]
        self.as_view = wraps(view_func)(paginator_view)

    @property
    def view_func_has_kwargs(self) -> bool:  # pragma: no cover
        return self.paginator.pass_parameter is not None

    def get_view_function(self) -> Callable:
        def as_view(controller: "ControllerBase", *args: Any, **kw: Any) -> Any:
            func_kwargs = dict(**kw)
            pagination_params = func_kwargs.pop(self.paginator_kwargs_name)
            if self.paginator.pass_parameter:
                func_kwargs[self.paginator.pass_parameter] = pagination_params

            items = self.view_func(controller, *args, **func_kwargs)
            assert (
                controller.context and controller.context.request
            ), "Request object is None"
            params = dict(kw)
            params["request"] = controller.context.request
            return self.paginator.paginate_queryset(items, **params)

        return as_view


class AsyncPaginatorOperation(PaginatorOperation):
    def get_view_function(self) -> Callable:
        async def as_view(controller: "ControllerBase", *args: Any, **kw: Any) -> Any:
            func_kwargs = dict(**kw)
            pagination_params = func_kwargs.pop(self.paginator_kwargs_name)
            if self.paginator.pass_parameter:
                func_kwargs[self.paginator.pass_parameter] = pagination_params

            items = await self.view_func(controller, *args, **func_kwargs)
            assert (
                controller.context and controller.context.request
            ), "Request object is None"

            params = dict(kw)
            params["request"] = controller.context.request
            paginate_queryset = cast(
                Callable, sync_to_async(self.paginator.paginate_queryset)
            )
            return await paginate_queryset(items, **params)

        return as_view
